import logging
import asyncio
from datetime import datetime
from minio import Minio
from src.writers.base import DataWriter
from typing import Dict
import io
import pyarrow as pa
import pyarrow.parquet as pq
from src.config.parser import Output

logger = logging.getLogger(__name__)

class S3Writer(DataWriter):
    def __init__(self, config: Output):
        logger.info("Initializing S3 writer...")
        self._init_config(config)
        logger.info(f"Initialized S3Writer with endpoint {self.endpoint_url}")

    def _init_config(self, config: Output) -> None:
        """Initialize S3 configuration"""
        # Format endpoint URL
        if not config.endpoint.startswith(('http://', 'https://')):
            self.endpoint_url = f"{'https://' if config.secure else 'http://'}{config.endpoint}"
            minio_endpoint = config.endpoint
        else:
            self.endpoint_url = config.endpoint
            minio_endpoint = config.endpoint.replace('http://', '').replace('https://', '')
            
        # Set bucket and path
        self.bucket = config.bucket
        self.s3_path = "minio-s3"  # Fixed subfolder for consistency
        
        # Initialize Minio client
        self.client = Minio(
            endpoint=minio_endpoint,
            access_key=config.access_key or 'minioadmin',
            secret_key=config.secret_key or 'minioadmin',
            secure=config.secure if config.secure is not None else True,
            region=config.region
        )
        
        # Ensure bucket exists
        if not self.client.bucket_exists(self.bucket):
            self.client.make_bucket(self.bucket)
            logger.info(f"Created bucket: {self.bucket}")

    async def push_data(self, data: Dict[str, pa.RecordBatch]) -> None:
        """Push data to S3"""
        try:
            # Process events
            for table_name in [t for t in data if t.endswith('_events')]:
                df = data[table_name].to_pandas()
                min_block = df['block_number'].min()
                max_block = df['block_number'].max()
                await self._write_parquet(table_name, pa.Table.from_pandas(df), min_block, max_block)

            # Process blocks
            blocks_data = self.combine_blocks(data)
            if blocks_data:
                df = blocks_data.to_pandas()
                min_block = df['block_number'].min()
                max_block = df['block_number'].max()
                await self._write_parquet('blocks', blocks_data, min_block, max_block)
                
        except Exception as e:
            logger.error(f"S3 write failed: {str(e)}")
            raise

    async def _write_parquet(self, table_name: str, data: pa.Table, min_block: int, max_block: int) -> None:
        """Write data to S3 as Parquet"""
        try:
            # Generate path
            timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
            subfolder = 'events/' + table_name.replace('_events', '') if table_name.endswith('_events') else 'blocks'
            key = f"{self.s3_path}/{subfolder}/{timestamp}_{min_block}_{max_block}.parquet"
            
            # Write to buffer
            buffer = io.BytesIO()
            pq.write_table(data, buffer)
            buffer.seek(0)
            size = buffer.getbuffer().nbytes
            
            # Upload to S3
            self.client.put_object(
                bucket_name=self.bucket,
                object_name=key,
                data=buffer,
                length=size
            )
            
            logger.info(f"Wrote {len(data)} rows to s3://{self.bucket}/{key}")
            
        except Exception as e:
            logger.error(f"Failed to write {table_name}: {str(e)}")
            raise
        finally:
            buffer.close() 


from ..writers.base import DataWriter
from ..config.parser import Writer, WriterKind
from ..writers.local_parquet import ParquetWriter

def create_writer(writer: Writer) -> DataWriter:
    match writer.kind:
        case WriterKind.LOCAL_PARQUET:
            return ParquetWriter(writer.config)
        case _:
            raise ValueError(f"Invalid writer kind: {writer.kind}")


import asyncio, logging
from typing import Dict
from ..writers.base import DataWriter
from ..config.parser import WriterConfig
import pyarrow as pa, pandas as pd
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from pathlib import Path

logger = logging.getLogger(__name__)

class ParquetWriter(DataWriter):
    def __init__(self, writer_config: WriterConfig):
        self._init_config(writer_config)
        logger.info(f"Initialized ParquetWriter with output path {self.path}")

    def _init_config(self, writer_config: WriterConfig) -> None:
        self.path = Path(writer_config.path)
        self.anchor_table = writer_config.anchor_table
        
        if not self.path:
            raise ValueError("output_path is required")          
        # Create output directory if it doesn't exist
        self.path.mkdir(parents=True, exist_ok=True)

    async def write_parquet(self, df: pd.DataFrame, path: str) -> None:
        """Write data to parquet file"""
        def write():
            logger.info(f"Writing {len(df)} rows to {path}")
            df.to_parquet(
                path,
                index=False,
                compression=None
            )
            logger.info(f"Successfully wrote {len(df)} rows to {path}")

        loop = asyncio.get_event_loop()
        with ThreadPoolExecutor() as pool:
            await loop.run_in_executor(pool, write)

    async def push_data(self, data: Dict[str, pa.RecordBatch]) -> None:
        """Push data to S3"""
        try:
            logger.info(f"Starting push_data for {len(data)} tables: {', '.join(data.keys())}")

            if self.anchor_table is None:
                await self._write_tables(data)
            else:
                non_anchor_tables = {k: v for k, v in data.items() if k != self.anchor_table}
                anchor_tables = {k: v for k, v in data.items() if k == self.anchor_table}
                await self._write_tables(non_anchor_tables)
                await self._write_tables(anchor_tables)
            
        except Exception as e:
            logger.error(f"Error writing to S3: {e}")
            raise

    async def _write_tables(self, data: Dict[str, pa.RecordBatch]) -> None:
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        
        event_tasks = {}
        for table_name, df in data.items():
            # Create table directory if it doesn't exist
            table_dir = self.path / table_name
            table_dir.mkdir(parents=True, exist_ok=True)
            
            event_tasks[table_name] = asyncio.create_task(
                self.write_parquet(
                    df=df.to_pandas(), 
                    path=table_dir / f"{table_name}_{timestamp}.parquet"
                ),
                name=f"write_{table_name}"
            )
        
        for name, task in event_tasks.items():
            try:
                await task
            except Exception as e:
                logger.error(f"Failed to write {name}: {str(e)}")
                raise Exception(f"Error writing parquet table into {name}: {e}")
            


import asyncio
import logging
from ..config.parser import parse_config, Pipeline, Provider, Step, StepKind
from ..utils.logging_setup import setup_logging
from typing import Dict, List
import copy
from dotenv import load_dotenv
from cherry_core.ingest import (
    start_stream,
    StreamConfig,
    Provider as CoreProvider,
    ProviderConfig as CoreProviderConfig,
    EvmQuery,
)
from cherry_core import evm_validate_block_data, evm_decode_events
import dacite
import pyarrow as pa
from ..writers.writer import create_writer

logger = logging.getLogger(__name__)

class Context:
    def __init__(self):
        self.steps = {}
    
    def add_step(self, kind: str, step: callable):
        self.steps[kind] = step

async def run_pipelines(path: str, context: Context):
    config = parse_config(path)

    tasks = {
        name: asyncio.create_task(
            run_pipeline(pipeline, context)
        )
        for name, pipeline in config.pipelines.items()
    }
    
    for name, task in tasks.items():
        try:
            await task
        except Exception as e:
            logger.error(f"Failed to run pipeline {name}: {str(e)}")
            raise Exception(f"Error running pipeline {name}: {e}")
        
def provider_to_stream_config(provider: Provider) -> StreamConfig:

    provider_config = CoreProviderConfig(
        url=provider.config.url,
        max_num_retries=provider.config.max_num_retries,
        retry_backoff_ms=provider.config.retry_backoff_ms,
        retry_base_ms=provider.config.retry_base_ms,
        retry_ceiling_ms=provider.config.retry_ceiling_ms,
        http_req_timeout_millis=provider.config.http_req_timeout_millis
    )

    core_provider = CoreProvider(
        kind=provider.kind,
        config=provider_config
    )

    return StreamConfig(
        format=provider.config.format,
        query=dacite.from_dict(data_class=EvmQuery, data=provider.config.query),
        provider=core_provider
    )

async def process_steps(res: Dict[str, pa.RecordBatch], steps: List[Step], context: Context) -> Dict[str, pa.RecordBatch]:
    logger.debug(f"Processing steps: {[step.kind for step in steps]}")
    logger.debug(f"Available custom steps: {list(context.steps.keys())}")
    
    # First process the built-in steps
    for step in steps:
        res = copy.deepcopy(res)
        if step.kind == StepKind.EVM_VALIDATE_BLOCK:
            evm_validate_block_data(blocks=res["blocks"], transactions=res["transactions"], logs=res["logs"], traces=res["traces"])
            pass
        elif step.kind == StepKind.EVM_DECODE_EVENTS:
            logger.debug(f"Processing decode events step: {step.config}")
            res[step.config['output_table']] = evm_decode_events(step.config['event_signature'], res[step.config['input_table']], step.config['allow_decode_fail'])
        elif context.steps[step.kind]:
                res = context.steps[step.kind](res, step.config)

    return res

async def run_pipeline(pipeline: Pipeline, context: Context):
    """Run a pipeline"""
    logger.info(f"Running pipeline: {pipeline.name}")

    stream_config = provider_to_stream_config(pipeline.provider)
    logger.info(stream_config)

    stream = start_stream(stream_config)

    writer = create_writer(pipeline.writer)

    while True:
        res = await stream.next()
        if res is None:
            break
            
        logger.info(f"Raw data num rows: {[record_batch.num_rows for _, record_batch in res.items()]}")
        res = await process_steps(res, pipeline.steps, context)
        logger.info(f"Processed data num rows: {[record_batch.num_rows for _, record_batch in res.items()]}")
        await writer.push_data(res)
        

        if res is None:
            break


import logging
import sys
from datetime import datetime
from pathlib import Path

# Global variables to track logging state
_is_logging_configured = False
_current_log_file = None

def setup_logging():
    """Configure logging for all modules"""
    global _is_logging_configured, _current_log_file
    
    # If logging is already configured, return the existing logger
    if _is_logging_configured:
        return logging.getLogger(__name__)

    # Create logs directory with absolute path if it doesn't exist
    log_dir = Path(__file__).parent.parent.parent / "logs"
    log_dir.mkdir(exist_ok=True)

    # Create a timestamp for the log file
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    _current_log_file = log_dir / f"blockchain_etl_{timestamp}.log"

    # Configure logging format
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    # File handler with UTF-8 encoding
    file_handler = logging.FileHandler(_current_log_file, encoding='utf-8')
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)

    # Console handler with proper encoding for Windows
    if sys.platform == 'win32':
        # Fix Windows console encoding
        sys.stdout.reconfigure(encoding='utf-8')
        console_handler = logging.StreamHandler(sys.stdout)
    else:
        console_handler = logging.StreamHandler(sys.stdout)
    
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(formatter)

    # Configure root logger
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)
    root_logger.addHandler(file_handler)
    root_logger.addHandler(console_handler)

    # Set higher log level for noisy third-party libraries
    logging.getLogger('botocore').setLevel(logging.WARNING)
    logging.getLogger('boto3').setLevel(logging.WARNING)
    logging.getLogger('urllib3').setLevel(logging.WARNING)
    logging.getLogger('s3transfer').setLevel(logging.WARNING)
    logging.getLogger('aiobotocore').setLevel(logging.WARNING)
    logging.getLogger('aioboto3').setLevel(logging.WARNING)
    logging.getLogger("awswrangler").setLevel(logging.WARNING)

    _is_logging_configured = True
    return root_logger

def get_current_log_file() -> Path:
    """Get the path to the current log file"""
    return _current_log_file 


import asyncio, boto3, logging, os
from typing import Optional, Dict, List
from writers.base import DataWriter
from config.parser import Output
import pyarrow as pa, pandas as pd
from concurrent.futures import ThreadPoolExecutor
from writers.writer import get_output_path
from datetime import datetime

logger = logging.getLogger(__name__)

def _get_athena_schema(record_batch: pa.RecordBatch) -> Dict[str, str]:
        """Convert PyArrow schema to Athena compatible types"""
        return {
            field.name: wr._data_types.pyarrow2athena(field.type)
            for field in record_batch.schema
        }

class AWSWranglerWriter(DataWriter):
    def __init__(self, config: Output):
        logger.info("Initializing AWS Wrangler S3 writer...")
        self._init_s3_config(config)
        
        if config.use_boto3:
            self._init_session(config)
        
        logger.info(f"Initialized AWSWranglerWriter with endpoint {self.endpoint_url}")

    def _init_session(self, config: Output) -> None:
        """Initialize AWS session"""
        
        self.session = boto3.Session(
            region_name=config.region or 'us-east-1'
        )
        
        # Configure AWS Wrangler
        wr.config.s3_endpoint_url = self.endpoint_url

    def _init_s3_config(self, config: Output) -> None:
        """Initialize S3 configuration"""
        # Format endpoint URL
        self.endpoint_url = config.endpoint
        self.s3_path = config.s3_path
            
        logger.info(f"Using S3 path: {self.s3_path}")
        
        self.partition_cols = config.partition_cols or {}
        self.default_partition_cols = config.default_partition_cols

    async def write_parquet(self, table_name: str, record_batch: pa.RecordBatch) -> None:
        """Write data to S3 as Parquet"""
        logger.info(f"Starting Parquet write for table {table_name}")
        
        table = pa.Table.from_batches([record_batch])
        pandas_df = table.to_pandas()

        schema = _get_athena_schema(record_batch)
        await self._write_to_s3(pandas_df, table_name, schema)

    async def _write_to_s3(self, df: pd.DataFrame, table_name: str, schema: Dict[str, str]) -> None:
        """Write DataFrame to S3 using thread pool"""
        def write():
            logger.info(f"Writing {len(df)} rows to {table_name}")
            wr.s3.to_parquet(
                df=df,
                boto3_session=self.session if self.session else None,
                path=self.s3_path,
                dataset=True,
                use_threads=True,
                mode="append",
                database=self.database,
                table=table_name,
                dtype=schema,
                partition_cols=self.partition_cols.get(table_name, self.default_partition_cols),
                schema_evolution=False,
            )
            logger.info(f"Successfully wrote {len(df)} rows to {self.s3_path}")

        loop = asyncio.get_event_loop()
        with ThreadPoolExecutor() as pool:
            await loop.run_in_executor(pool, write)

    async def push_data(self, data: Dict[str, pa.RecordBatch]) -> None:
        """Push data to S3"""
        try:
            logger.info(f"Starting push_data for {len(data)} tables: {', '.join(data.keys())}")

            if self.anchor_table is None:
                await self._write_tables(data)
            else:
                await self._write_tables(filter(lambda x, _: x != self.anchor_table, data))
                await self._write_tables(filter(lambda x, _: x == self.anchor_table, data))
            
        except Exception as e:
            logger.error(f"Error writing to S3: {e}")
            raise

    async def _write_tables(self, data: Dict[str, pa.RecordBatch]) -> None:
        """Write event data to S3"""
        event_tasks = {
            name: asyncio.create_task(self.write_parquet(name, df), name=f"write_{name}")
            for name, df in data.items()
        }
        
        for name, task in event_tasks.items():
            try:
                await task
            except Exception as e:
                logger.error(f"Failed to write {name}: {str(e)}")
                raise Exception(f"Error writing parquet table into {name}: {e}")

    def _get_s3_path(self, table: str, min_block: int, max_block: int) -> str:
        """Generate S3 path for the table"""
        timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
        base_path = f"s3://{self.bucket}/{self.s3_path}"
        
        if table.endswith('_events'):
            event_type = table.replace('_events', '')
            return f"{base_path}/events/{event_type}/{timestamp}_{min_block}_{max_block}.parquet"
        else:
            return f"{base_path}/blocks/{timestamp}_{min_block}_{max_block}.parquet"



from sqlalchemy import Engine, text
import logging
import pyarrow as pa
import psycopg2
from psycopg2.extensions import connection as pg_connection
from src.schemas.blockchain_schemas import BLOCKS, TRANSACTIONS, EVENTS
from src.schemas.base import SchemaConverter
from src.types.data import Data
from src.writers.base import DataWriter
import polars as pl
from typing import Tuple, Optional, Dict, List, Any
import asyncio
from src.config.parser import Output
from sqlalchemy import create_engine
import pandas as pd
from src.schemas.blockchain_schemas import BlockchainSchema
import time
from io import StringIO
import csv

logger = logging.getLogger(__name__)

class PostgresWriter(DataWriter):
    BLOCKS_TABLE = BlockchainSchema.BLOCKS_TABLE_NAME

    def __init__(self, config: Output):
        self.host = config.host
        self.port = config.port
        self.database = config.database
        self.username = config.username
        self.password = config.password
        self.conn = None
        self.initialized = False
        self._init_task = asyncio.create_task(self._initialize())

    def _connect(self, database: str = None) -> pg_connection:
        """Establish database connection"""
        try:
            conn = psycopg2.connect(
                host=self.host,
                port=self.port,
                user=self.username,
                password=self.password,
                database=database or self.database,
                connect_timeout=3
            )
            conn.autocommit = True
            return conn
        except Exception as e:
            raise ConnectionError(f"Failed to connect to PostgreSQL: {str(e)}")

    async def _initialize(self) -> None:
        """Initialize database and tables"""
        try:
            try:
                self.conn = self._connect()
            except Exception:
                # Create database if it doesn't exist
                sys_conn = self._connect('postgres')
                with sys_conn.cursor() as cur:
                    cur.execute(f'CREATE DATABASE {self.database}')
                sys_conn.close()
                self.conn = self._connect()

            # Create tables
            with self.conn.cursor() as cur:
                cur.execute(BlockchainSchema.get_postgres_blocks_ddl())
                cur.execute(BlockchainSchema.get_postgres_events_ddl('approval_events'))
                cur.execute(BlockchainSchema.get_postgres_events_ddl('transfer_events'))
            
            self.initialized = True
            logger.info(f"Initialized PostgreSQL connection to {self.host}:{self.port}")
        except Exception as e:
            logger.error(f"Failed to initialize PostgreSQL: {str(e)}")
            raise

    def _prepare_data(self, batch: pa.RecordBatch) -> pd.DataFrame:
        """Convert Arrow batch to DataFrame with proper types"""
        df = batch.to_pandas()
        
        # Convert timestamps
        if 'block_timestamp' in df.columns:
            df['block_timestamp'] = pd.to_datetime(
                df['block_timestamp'].apply(lambda x: int(x[2:], 16) if isinstance(x, str) and x.startswith('0x') else x),
                unit='s'
            )
        
        # Convert numeric columns
        for col in ['log_index', 'transaction_index', 'block_number']:
            if col in df.columns:
                df[col] = pd.to_numeric(df[col], errors='coerce')
        
        return df

    async def _insert_data(self, table_name: str, df: pd.DataFrame) -> None:
        """Insert data using COPY command"""
        if df.empty:
            return

        output = StringIO()
        writer = csv.writer(output, delimiter='\t', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        writer.writerows(df.values.tolist())
        output.seek(0)

        with self.conn.cursor() as cur:
            if table_name == 'blocks':
                # Use temp table for blocks to handle upserts
                temp_table = f"temp_blocks_{int(time.time())}"
                cur.execute(f"CREATE TEMP TABLE {temp_table} (LIKE blocks INCLUDING ALL)")
                cur.copy_expert(f"COPY {temp_table} ({','.join(df.columns)}) FROM STDIN WITH CSV DELIMITER E'\\t' QUOTE '\"'", output)
                
                cur.execute(f"""
                    INSERT INTO blocks 
                    SELECT * FROM {temp_table}
                    ON CONFLICT (block_number) 
                    DO UPDATE SET 
                        block_timestamp = EXCLUDED.block_timestamp,
                        block_hash = EXCLUDED.block_hash,
                        block_parent_hash = EXCLUDED.block_parent_hash,
                        block_transaction_count = EXCLUDED.block_transaction_count,
                        block_size = EXCLUDED.block_size
                """)
                cur.execute(f"DROP TABLE {temp_table}")
            else:
                cur.copy_expert(f"COPY {table_name} ({','.join(df.columns)}) FROM STDIN WITH CSV DELIMITER E'\\t' QUOTE '\"'", output)

        output.close()
        logger.info(f"Inserted {len(df)} rows into {table_name}")

    async def push_data(self, data: dict[str, pa.RecordBatch]) -> None:
        """Push data to PostgreSQL"""
        if not self.initialized:
            await self._init_task

        # Process events
        for table_name in [name for name in data if name.endswith('_events')]:
            df = self._prepare_data(data[table_name])
            await self._insert_data(table_name, df)

        # Process blocks
        blocks_tables = [name for name in data if name.startswith('blocks_')]
        if blocks_tables:
            blocks_dfs = [self._prepare_data(data[table]) for table in blocks_tables]
            combined_blocks = pd.concat(blocks_dfs).drop_duplicates(subset=['block_number'])
            await self._insert_data('blocks', combined_blocks) 


from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, List
import polars as pl
import logging
import pyarrow as pa

logger = logging.getLogger(__name__)

class DataWriter(ABC):
    """Base class for data writers"""

    @abstractmethod
    async def push_data(self, data: Dict[str, pa.RecordBatch]) -> None:
        """Push data to target storage"""
        pass

